import numpy as np
import unittest
import ica
import time


def find_sources_order(S_true, S_estimated):
    NCOMP, NVOX = S_true.shape
    C = np.corrcoef(S_true, S_estimated)
    C = np.abs(C[NCOMP:, :NCOMP])
    idx = np.argmax(C, axis=0)
    return idx


def mean_corr(S_true, S_estimated):
    # Always correlate the smallest dimension
    if S_true.shape[0] > S_true.shape[1]:
        C = np.corrcoef(S_true.T, S_estimated.T)
    else:
        C = np.corrcoef(S_true, S_estimated)
    NCOMP = min(S_true.shape)
    C = np.abs(C[NCOMP:, :NCOMP])
    return np.diag(C).mean()


def auto_cov(A):
    A = A - A.mean(axis=1).reshape((-1, 1))
    cov = np.dot(A, A.T) / (A.shape[1] - 1)
    return(cov)


class test_ica_methods(unittest.TestCase):

    def setUp(self):

        self.NCOMP = 100
        self.NVOX = 50000
        self.NSUB = 500
        # Setting up X = A @ S
        self.S = np.random.logistic(0, 1, (self.NCOMP, self.NVOX))
        self.A = np.random.normal(0, 1, (self.NSUB, self.NCOMP))
        self.clean_data = np.dot(self.A, self.S)
        self.clean_data = self.clean_data - \
            self.clean_data.mean(axis=1).reshape((-1, 1))
        # self.clean_data = self.clean_data - self.clean_data.mean(axis=0)
        self.noisy_data = self.clean_data + \
            np.random.normal(0, 1, self.clean_data.shape)
        self.noisy_data = self.noisy_data - \
            self.noisy_data.mean(axis=1).reshape((-1, 1))
        # self.noisy_data = self.noisy_data - self.clean_data.mean(axis=0)

    def test_PCA_whitening_clean(self):
        start = time.time()
        x_white, white, dewhite = ica.pca_whiten(self.clean_data, self.NCOMP)
        end = time.time()
        print('\tPCA whitening time: {:.2f} seconds'.format(end - start))
        # Check output dimensions
        self.assertEqual(x_white.shape, (self.NCOMP, self.NVOX))
        self.assertEqual(white.shape, (self.NCOMP, self.NSUB))
        self.assertEqual(dewhite.shape, (self.NSUB, self.NCOMP))

        # Check variance is 1
        var = x_white.var(axis=1)
        self.assertLess(np.linalg.norm(var - 1.0), 1e-2)

        # Test wether the covariance of x_white is the identity
        cov = auto_cov(x_white)
        self.assertLess(np.linalg.norm(cov - np.eye(self.NCOMP)
                                       ) / self.NCOMP / self.NCOMP, 1e-6)
        # Test wether white and dewhite are orthogonals
        eye = np.dot(white, dewhite)
        self.assertLess(np.linalg.norm(eye - np.eye(self.NCOMP)
                                       ) / self.NCOMP / self.NCOMP, 1e-4)
        eye = np.dot(dewhite, white)
        self.assertLess(np.linalg.norm(eye - np.eye(self.NSUB)
                                       ) / self.NSUB / self.NSUB, 1e-4)

    # @unittest.skip("PCAwhiten not passing")
    def test_PCA_whitening_noisy(self):
        start = time.time()
        x_white, white, dewhite = ica.pca_whiten(self.noisy_data, self.NCOMP)
        end = time.time()
        print('\ttime: {:.2f} seconds'.format(end - start))
        self.assertEqual(x_white.shape, (self.NCOMP, self.NVOX))
        self.assertEqual(white.shape, (self.NCOMP, self.NSUB))
        self.assertEqual(dewhite.shape, (self.NSUB, self.NCOMP))

        cov = auto_cov(x_white)
        self.assertLess(np.linalg.norm(cov - np.eye(self.NCOMP)
                                       ) / self.NCOMP / self.NCOMP, 1e-6)

        # Test wether white and dewhite are orthogonals
        eye = np.dot(white, dewhite)
        self.assertLess(np.linalg.norm(eye - np.eye(self.NCOMP)
                                       ) / self.NCOMP / self.NCOMP, 1e-4)
        eye = np.dot(dewhite, white)
        self.assertLess(np.linalg.norm(eye - np.eye(self.NSUB)
                                       ) / self.NSUB / self.NSUB, 1e-4)

    # @unittest.skip("PCAwhiten not passing")
    def test_ICA_infomax_clean(self):

        start = time.time()
        A, S, _ = ica.ica1(self.clean_data, self.NCOMP)
        end = time.time()
        print('\ttime: {}:.2f'.format(end - start))

        # Check right dimensions of Output
        self.assertEqual(A.shape, (self.NSUB, self.NCOMP))
        self.assertEqual(S.shape, (self.NCOMP, self.NVOX))

        idx = find_sources_order(self.S, S)
        S = S[idx, :]
        A = A[:, idx]
        # Check the accuracy of output
        self.assertGreater(mean_corr(self.S, S), 0.95)
        self.assertGreater(mean_corr(self.A, A), 0.95)


if __name__ == '__main__':
    unittest.main()



'''
Independent Component Analysis (ICA):
This script computes ICA using the INFOMAX criteria.
The preprocessing steps include demeaning and whitening.
'''
import numpy as np
from numpy import dot
from numpy.linalg import matrix_rank, inv
from numpy.random import permutation
from scipy.linalg import eigh
from scipy.linalg import norm as mnorm

# Global constants
EPS = 1e-16
MAX_W = 1e8
ANNEAL = 0.9
MAX_STEP = 500
MIN_LRATE = 1e-6
W_STOP = 1e-6


def norm(x):
    """Computes the norm of a vector or the Frobenius norm of a
    matrix_rank

    """
    return mnorm(x.ravel())


class ica:

    def __init__(self, n_components=10):
        self.n_comp = n_components

    def fit(self, x2d):
        x_white, self.white, self.dewhite\
            = pca_whiten(x2d, self.ncomp)
        self.mix, self.sources, self.unmix\
            = infomax1(x_white, self.n_comp)
        return self


def diagsqrts(w):
    """
    Returns direct and inverse square root normalization matrices
    """
    Di = np.diag(1. / (np.sqrt(w) + np.finfo(float).eps))
    D = np.diag(np.sqrt(w))
    return D, Di


def pca_whiten(x2d, n_comp, verbose=True):
    """ data Whitening
    *Input
    x2d : 2d data matrix of observations by variables
    n_comp: Number of components to retain
    *Output
    Xwhite : Whitened X
    white : whitening matrix (Xwhite = np.dot(white,X))
    dewhite : dewhitening matrix (X = np.dot(dewhite,Xwhite))
    """
    x2d_demean = x2d - x2d.mean(axis=1).reshape((-1, 1))
    NSUB, NVOX = x2d_demean.shape
    if NSUB > NVOX:
        cov = dot(x2d_demean.T, x2d_demean) / (NSUB - 1)
        w, v = eigh(cov, eigvals=(NVOX - n_comp, NVOX - 1))
        D, Di = diagsqrts(w)
        u = dot(dot(x2d_demean, v), Di)
        x_white = v.T
        white = dot(Di, u.T)
        dewhite = dot(u, D)
    else:
        cov = dot(x2d_demean, x2d_demean.T) / (NVOX - 1)
        w, u = eigh(cov, eigvals=(NSUB - n_comp, NSUB - 1))
        D, Di = diagsqrts(w)
        white = dot(Di, u.T)
        x_white = dot(white, x2d_demean)
        dewhite = dot(u, D)
    return (x_white, white, dewhite)


def w_update(weights, x_white, bias1, lrate1):
    """ Update rule for infomax
    This function recieves parameters to update W1
    * Input
    W1: unmixing matrix (must be a square matrix)
    Xwhite1: whitened data
    bias1: current estimated bias
    lrate1: current learning rate
    startW1: in case update blows up it will start again from startW1
    * Output
    W1: updated mixing matrix
    bias: updated bias
    lrate1: updated learning rate
    """
    NCOMP, NVOX = x_white.shape
    block1 = int(np.floor(np.sqrt(NVOX / 3)))
    permute1 = permutation(NVOX)
    for start in range(0, NVOX, block1):
        if start + block1 < NVOX:
            tt2 = start + block1
        else:
            tt2 = NVOX
            block1 = NVOX - start

        unmixed = dot(weights, x_white[:, permute1[start:tt2]]) + bias1
        logit = 1 - (2 / (1 + np.exp(-unmixed)))
        weights = weights + lrate1 * dot(block1 * np.eye(NCOMP) +
                                         dot(logit, unmixed.T), weights)
        bias1 = bias1 + lrate1 * logit.sum(axis=1).reshape(bias1.shape)
        # Checking if W blows up
        if (np.isnan(weights)).any() or np.max(np.abs(weights)) > MAX_W:
            print("Numeric error! restarting with lower learning rate")
            lrate1 = lrate1 * ANNEAL
            weights = np.eye(NCOMP)
            bias1 = np.zeros((NCOMP, 1))
            error = 1

            if lrate1 > 1e-6 and \
               matrix_rank(x_white) < NCOMP:
                print("Data 1 is rank defficient"
                      ". I cannot compute " +
                      str(NCOMP) + " components.")
                return (None, None, None, 1)

            if lrate1 < 1e-6:
                print("Weight matrix may"
                      " not be invertible...")
                return (None, None, None, 1)
            break
        else:
            error = 0

    return(weights, bias1, lrate1, error)

# infomax1: single modality infomax


def infomax1(x_white, verbose=False):
    """Computes ICA infomax in whitened data
    Decomposes x_white as x_white=AS
    *Input
    x_white: whitened data (Use PCAwhiten)
    verbose: flag to print optimization updates
    *Output
    A : mixing matrix
    S : source matrix
    W : unmixing matrix
    """
    NCOMP = x_white.shape[0]
    # Initialization
    weights = np.eye(NCOMP)
    old_weights = np.eye(NCOMP)
    d_weigths = np.zeros(NCOMP)
    old_d_weights = np.zeros(NCOMP)
    lrate = 0.005 / np.log(NCOMP)
    bias = np.zeros((NCOMP, 1))
    change = 1
    angle_delta = 0
    if verbose:
        print("Beginning ICA training...")
    step = 1

    while step < MAX_STEP and change > W_STOP:

        (weights, bias, lrate, error) = w_update(weights, x_white, bias, lrate)

        if error != 0:
            step = 1
            error = 0
            lrate = lrate * ANNEAL
            weights = np.eye(NCOMP)
            old_weights = np.eye(NCOMP)
            d_weigths = np.zeros(NCOMP)
            old_d_weights = np.zeros(NCOMP)
            bias = np.zeros((NCOMP, 1))
        else:
            d_weigths = weights - old_weights
            change = norm(d_weigths)**2

            if step > 2:
                angle_delta = np.arccos(
                    np.sum(d_weigths * old_d_weights) /
                    (norm(d_weigths) * norm(old_d_weights) + 1e-8)
                ) * 180 / np.pi

            old_weights = np.copy(weights)

            if angle_delta > 60:
                lrate = lrate * ANNEAL
                old_d_weights = np.copy(d_weigths)
            elif step == 1:
                old_d_weights = np.copy(d_weigths)

            if verbose and change < W_STOP:
                print("Step %d: Lrate %.1e,"
                      "Wchange %.1e,"
                      "Angle %.2f" % (step, lrate,
                                      change, angle_delta))

        step = step + 1

    # A,S,W
    return (inv(weights), dot(weights, x_white), weights)

# Single modality ICA


def ica1(x_raw, ncomp, verbose=False):
    '''
    Single modality Independent Component Analysis
    '''
    if verbose:
        print("Whitening data...")
    x_white, _, dewhite = pca_whiten(x_raw, ncomp)
    if verbose:
        print('x_white shape: %d, %d' % x_white.shape)
        print("Done.")
    if verbose:
        print("Running INFOMAX-ICA ...")
    mixer, sources, unmixer = infomax1(x_white, verbose)
    mixer = dot(dewhite, mixer)

    scale = sources.std(axis=1).reshape((-1, 1))
    sources = sources / scale
    scale = scale.reshape((1, -1))
    mixer = mixer * scale

    if verbose:
        print("Done.")
    return (mixer, sources, unmixer)


def icax(x_raw, ncomp, verbose=True):

    if verbose:
        print("Whitening data...")
    x_white, _, dewhite = pca_whiten(x_raw, ncomp)

    mixer_list = []
    sources_list = []
    for it in range(10):
        if verbose:
            print('Run number %d' % it)
            print("Running INFOMAX-ICA ...")
        mixer, sources, unmix = infomax1(x_white, verbose)
        mixer_list.append(mixer)
        sources_list.append(sources)

    # Reorder all sources to the order of the first
    S1 = sources_list[0]
    for it in range(1, 10):
        S2 = sources_list[it]
        A2 = mixer_list[it]
        cor_m = np.corrcoef(S1, S2)[:ncomp, ncomp:]
        idx = np.argmax(np.abs(cor_m), axis=1)
        S2 = S2[idx, :]
        A2 = A2[:, idx]
        cor_m = np.corrcoef(S1, S2)[:ncomp, ncomp:]
        S2 = S2 * np.sign(np.diag(cor_m)).reshape((ncomp, 1))
        A2 = A2 * np.sign(np.diag(cor_m)).reshape((1, ncomp))
        sources_list[it] = S2
        mixer_list[it] = A2

    # Average sources
    temp_sources = np.zeros(sources.shape)
    temp_mixer = np.zeros(mixer.shape)
    for sources, mixer in zip(sources_list, mixer_list):
        temp_sources = temp_sources + sources
        temp_mixer = temp_mixer + mixer

    temp_sources = temp_sources / 10.0
    temp_mixer = temp_mixer / 10.0

    return (temp_mixer, temp_sources)



# Import ica function
from ica import ica1
import numpy as np
import matplotlib.pyplot as plt
import time

def main():
    # Define matrix dimensions
    Nobs = 1000 # Number of observation
    Nvars = 50000 # Number of variables
    Ncomp = 100 # Number of components

    # Simulated true sources
    S_true = np.random.logistic(0,1,(Ncomp,Nvars))
    # Simulated true mixing
    A_true = np.random.normal(0,1,(Nobs,Ncomp))
    # X = AS
    X = np.dot(A_true,S_true)
    # add some noise
    X = X + np.random.normal(0,1,X.shape)
    # apply ICA on X and ask for 2 components

    model = ica1(Ncomp)
    
    start = time.time()
    A,S = model.fit(X)
    total = time.time() - start
    print('total time: {}'.format(total))
    # compare if our estimates are accurate
    # correlate A with Atrue and take 
    aCorr = np.abs(np.corrcoef(A.T,A_true.T)[:Ncomp,Ncomp:]).max(axis = 0).mean()
    sCorr = np.abs(np.corrcoef(S,S_true)[:Ncomp,Ncomp:]).max(axis = 0).mean()

    print("Accuracy of estimated sources: %.2f"%sCorr)
    print("Accuracy of estimated mixing: %.2f"%aCorr)

if __name__=="__main__":

    import theano.sandbox.cuda
    theano.sandbox.cuda.use('gpu')

    main()



'''
Independent Component Analysis (ICA):
This script computes ICA using the INFOMAX criteria.
The preprocessing steps include demeaning and whitening.
'''
import numpy as np
from numpy import dot
from numpy.linalg import matrix_rank, inv
from numpy.random import permutation
from scipy.linalg import eigh

# Theano Imports
import theano.tensor as T
import theano
from theano import shared

# Global constants
EPS = 1e-18
MAX_W = 1e8
ANNEAL = 0.9
MAX_STEP = 500
MIN_LRATE = 1e-6
W_STOP = 1e-6


class ica_gpu(object):
    """
    Infomax ICA for one data modality
    """
    def __init__(self, n_comp=10, verbose=False):

        # Theano initialization
        self.T_weights = shared(np.eye(n_comp, dtype=np.float32))
        self.T_bias = shared(np.ones((n_comp, 1), dtype=np.float32))

        T_p_x_white = T.fmatrix()
        T_lrate = T.fscalar()
        T_block = T.fscalar()
        T_unmixed = T.dot(self.T_weights,T_p_x_white) + T.addbroadcast(self.T_bias,1)
        T_logit = 1 - 2 / (1 + T.exp(-T_unmixed))

        T_out =  self.T_weights +  T_lrate * T.dot(T_block * T.identity_like(self.T_weights) + T.dot(T_logit, T.transpose(T_unmixed)), self.T_weights)
        T_bias_out = self.T_bias + T_lrate * T.reshape(T_logit.sum(axis=1), (-1,1))
        T_max_w = T.max(self.T_weights)
        T_isnan = T.any(T.isnan(self.T_weights))

        self.w_up_fun = theano.function([T_p_x_white, T_lrate, T_block],
                                        [T_max_w, T_isnan],
                                        updates=[(self.T_weights, T_out),
                                                 (self.T_bias, T_bias_out)],
                                        allow_input_downcast=True)

        T_matrix = T.fmatrix()
        T_cov = T.dot(T_matrix,T.transpose(T_matrix))/T_block
        self.cov_fun = theano.function([T_matrix, T_block], T_cov, allow_input_downcast=True)
        
        self.loading = None
        self.sources = None
        self.weights = None
        self.n_comp = n_comp
        self.verbose = verbose

    def __pca_whiten(self, x2d):
        """ data Whitening
        *Input
        x2d : 2d data matrix of observations by variables
        n_comp: Number of components to retain
        *Output
        Xwhite : Whitened X
        white : whitening matrix (Xwhite = np.dot(white,X))
        dewhite : dewhitening matrix (X = np.dot(dewhite,Xwhite))
        """
        NSUB, NVOX = x2d.shape
        x2d_demean = x2d - x2d.mean(axis=1).reshape((-1,1))
        #cov = dot(x2d_demean, x2d_demean.T) / ( NVOX -1 )
        cov = self.cov_fun(x2d_demean, NVOX-1)
        w, v = eigh(cov,eigvals=(NSUB-self.n_comp,NSUB-1))
        D = np.diag(1./(np.sqrt(w ) ))
        white = dot(D,v.T)
        D = np.diag( np.sqrt(w))
        dewhite = dot(v,D)
        x_white = dot(white,x2d_demean)
        return (x_white, white, dewhite)

    def __w_update(self, x_white, lrate1):
        """ Update rule for infomax
        This function recieves parameters to update W1
        * Input
        W1: unmixing matrix (must be a square matrix)
        Xwhite1: whitened data
        bias1: current estimated bias
        lrate1: current learning rate
        startW1: in case update blows up it will start again from startW1
        * Output
        W1: updated mixing matrix
        bias: updated bias
        lrate1: updated learning rate
        """
        error = 0
        NVOX = x_white.shape[1]
        NCOMP = x_white.shape[0]
        block1 = int(np.floor(np.sqrt(NVOX / 3)))
        permute1 = permutation(NVOX)
        p_x_white = x_white[:, permute1].astype(np.float32)

        for start in range(0, NVOX, block1):
            if start + block1 < NVOX:
                tt2 = start + block1
            else:
                tt2 = NVOX
                block1 = NVOX - start

            max_w, isnan = self.w_up_fun(p_x_white[:,start:tt2], lrate1, block1)
                
                # Checking if W blows up
            if isnan or max_w > MAX_W:
                print("Numeric error! restarting with lower learning rate")
                lrate1 = lrate1 * ANNEAL
                self.T_weights.set_value(np.eye(NCOMP, dtype=np.float32))
                self.T_bias.set_value( np.zeros((NCOMP, 1), dtype=np.float32))
                error = 1

                if lrate1 > 1e-6 and \
                   matrix_rank(x_white) < NCOMP:
                    print("Data 1 is rank defficient"
                          ". I cannot compute " +
                          str(NCOMP) + " components.")
                    return (0, 1)

                if lrate1 < 1e-6:
                    print("Weight matrix may"
                          " not be invertible...")
                    return (0, 1)
        
        return(lrate1, error)

    def __infomax(self, x_white):
        """Computes ICA infomax in whitened data
        Decomposes x_white as x_white=AS
        *Input
        x_white: whitened data (Use PCAwhiten)
        verbose: flag to print optimization updates
        *Output
        A : mixing matrix
        S : source matrix
        W : unmixing matrix
        """
    
        NCOMP = self.n_comp    
        # Initialization
        self.T_weights.set_value( np.eye(NCOMP, dtype=np.float32))
        weights = np.eye(NCOMP)
        old_weights = np.eye(NCOMP)
        d_weigths = np.zeros(NCOMP)
        old_d_weights = np.zeros(NCOMP)
        lrate = 0.005 / np.log(NCOMP)
        self.T_bias.set_value(np.zeros((NCOMP, 1), dtype=np.float32))
        change = 1
        angle_delta = 0
        step = 1
        if self.verbose:
            print("Beginning ICA training...")
        
            
        while step < MAX_STEP and change > W_STOP:

            (lrate, error) = self.__w_update( x_white, lrate)
            
            if error != 0:
                step = 1
                error = 0
                lrate = lrate * ANNEAL
                self.T_weights.set_value( np.eye(NCOMP, dtype=np.float32))
                old_weights = np.eye(NCOMP)
                d_weigths = np.zeros(NCOMP)
                old_d_weights = np.zeros(NCOMP)
                self.T_bias.set_value(np.zeros((NCOMP, 1), dtype=np.float32))
            else:
                weights = self.T_weights.get_value()
                d_weigths = weights - old_weights
                change = np.linalg.norm(d_weigths, 'fro')**2

                if step > 2:
                    angle_delta = np.arccos(np.sum(d_weigths * old_d_weights) /
                                            (np.linalg.norm(d_weigths, 'fro')) /
                                            (np.linalg.norm(old_d_weights, 'fro')))
                    angle_delta = angle_delta * 180 / np.pi

                old_weights = np.copy(weights)

                if angle_delta > 60:
                    lrate = lrate * ANNEAL
                    old_d_weights = np.copy(d_weigths)
                elif step == 1:
                    old_d_weights = np.copy(d_weigths)

                if (self.verbose and step % 10 == 0) or\
                (self.verbose and change < W_STOP):
                    print("Step %d: Lrate %.1e,"
                          "Wchange %.1e,"
                          "Angle %.2f" % (step, lrate,
                                          change, angle_delta))

            step = step + 1

        # A,S,W
        return (inv(weights), dot(weights, x_white), weights)

    def fit(self, x_raw):
        '''
        Single modality Independent Component Analysis
        '''
        if self.verbose:
            print("Whitening data...")
        x_white, _, dewhite = self.__pca_whiten(x_raw)
        if self.verbose:
            print("Done.")
            print("Running INFOMAX-ICA ...")
        loading, self.sources, self.weights = self.__infomax(x_white)

        self.loading = dot(dewhite, loading)
        if self.verbose:
            print("Done.")
        return (self.loading, self.sources)













from distutils.core import setup

setup(
    name='ica',
    packages=['ica'],
    description='Independent component analysis with INFOMAX algorithm',
    version='0.4',
    url='https://github.com/alvarouc/ica/',
    author='Alvaro Ulloa',
    author_email='alvarouc@gmail.com',
    download_url='https://github.com/alvarouc/ica/tarball/0.4',
    keywords=['infomax', 'ica', 'independent component analysis'],
    install_requires=[
        'numpy',
        'scipy',
    ],
)
